@@ -1,18 +1,19 @@
 #include "vadr.h"
 
-SEXP stringify_item(SEXP, char *);
-int sprintdouble(char *, double);
+SEXP stringify_item(SEXP, char *, char *);
+int snprintdouble(char *, size_t, double);
+SEXP weakref_(SEXP, SEXP);
 
 #if defined(__STDC_VERSION__) && __STDC_VERSION__ >= 199901L && !defined(__MINGW32__)
 
-int sprintdouble(char *buf, double arg) {
-  return sprintf(buf, "%la", arg);
+int snprintdouble(char *buf, size_t n, double arg) {
+  return snprintf(buf, n, "%la", arg);
 }
 
 #else
 
 /* hack for stdlibs that don't support "%a" printf conversion */
-int sprintdouble(char *buf, double arg) {
+int snprintdouble(char *buf, size_t n, double arg) {
   int written = 0;
   union {
     double f;
@@ -21,9 +22,10 @@ int sprintdouble(char *buf, double arg) {
   data.f = arg;
 
   for (int i = 0; i < sizeof(double) / sizeof(char); i++) {
-    int chars = sprintf(buf, "%02x", data.ch[i]);
+    int chars = snprintf(buf, n, "%02x", data.ch[i]);
     written += chars;
     buf += chars;
+    n -= chars;
   }
   return written;
 }
@@ -67,15 +69,16 @@ SEXP _string_reps(SEXP list) {
        "c0x0123456789abcdef=c_0x0123456789abcdef/0x0123456789abcdef/0x01234567abcdef/0x01234567abcdef" = 97 chars*/
     char buf[128];
     char *bufptr = buf;
+    char *end = bufptr + sizeof(buf) - 1;
 
     if (in_names != R_NilValue) {
       SEXP name = STRING_ELT(in_names, i);
       if (name != R_BlankString) {
-        bufptr += sprintf(bufptr, "c%p=", R_CHAR(name));
+        bufptr += snprintf(bufptr, end-bufptr, "c%p=", R_CHAR(name));
       }
     }
 
-    stringify_item(item, bufptr);
+    stringify_item(item, bufptr, end);
     SET_STRING_ELT(out_reps, i, mkChar(buf));
   }
 
@@ -91,7 +94,7 @@ SEXP _string_reps(SEXP list) {
 /* Construct a string identifying some SEXP, either as a scalar value or as a pointer.
    If we use its pointer, mark the item immutable.
    Return that pointer, or R_NilValue. */
-SEXP stringify_item(SEXP item, char *bufptr) {
+SEXP stringify_item(SEXP item, char *bufptr, char* end) {
   int done = 0;
   PROTECT_INDEX ix;
   SEXP item_ptr;
@@ -105,7 +108,7 @@ SEXP stringify_item(SEXP item, char *bufptr) {
     case CHARSXP:
       /* interned string, represent its pointer */
       REPROTECT(item_ptr = item, ix);
-      bufptr += sprintf(bufptr, "c%p", CHAR(item_ptr));
+      bufptr += snprintf(bufptr, end-bufptr, "c%p", CHAR(item_ptr));
       done = 1;
       break;
     case REALSXP:
@@ -116,40 +119,40 @@ SEXP stringify_item(SEXP item, char *bufptr) {
          and don't hold a ref to a scalar. */
       if (LENGTH(item) == 0) {
         switch(TYPEOF(item)) {
-        case REALSXP: bufptr += sprintf(bufptr, "r0"); break;
-        case INTSXP: bufptr +=  sprintf(bufptr, "i0"); break;
-        case LGLSXP: bufptr += sprintf(bufptr, "l0"); break;
-        case STRSXP: bufptr += sprintf(bufptr, "s0"); break;
+        case REALSXP: bufptr += snprintf(bufptr, end-bufptr, "r0"); break;
+        case INTSXP: bufptr +=  snprintf(bufptr, end-bufptr, "i0"); break;
+        case LGLSXP: bufptr += snprintf(bufptr, end-bufptr, "l0"); break;
+        case STRSXP: bufptr += snprintf(bufptr, end-bufptr, "s0"); break;
         default: error("Unexpected type %s (this shouldn't happen)", TYPEOF(item));
         }
       } else if (LENGTH(item) == 1) {
         switch(TYPEOF(item)) {
         case REALSXP:
-          bufptr += sprintf(bufptr, "r");
-          bufptr += sprintdouble(bufptr, REAL(item)[0]);
+          bufptr += snprintf(bufptr, end-bufptr, "r");
+          bufptr += snprintdouble(bufptr, end-bufptr, REAL(item)[0]);
           break;
-        case INTSXP: bufptr += sprintf(bufptr, "i%x", INTEGER(item)[0]); break;
-        case LGLSXP: bufptr += sprintf(bufptr, "l%x", LOGICAL(item)[0]); break;
+        case INTSXP: bufptr += snprintf(bufptr, end-bufptr, "i%x", INTEGER(item)[0]); break;
+        case LGLSXP: bufptr += snprintf(bufptr, end-bufptr, "l%x", LOGICAL(item)[0]); break;
         case STRSXP:
           REPROTECT(item_ptr = STRING_ELT(item, 0), ix);
-          bufptr += sprintf(bufptr, "s%p", CHAR(item_ptr)); break;
+          bufptr += snprintf(bufptr, end-bufptr, "s%p", CHAR(item_ptr)); break;
         default: error("Unexpected type %s (this shouldn't happen)", TYPEOF(item));
         }
       } else {
         /* for non-scalar vectors, represent the pointer */
         REPROTECT(item_ptr = item, ix);
-        bufptr += sprintf(bufptr, "v%p", (void *)item_ptr);
+        bufptr += snprintf(bufptr, end-bufptr, "v%p", (void *)item_ptr);
       }
       done = 1;
       break;
     case VECSXP:
       REPROTECT(item_ptr = item, ix);
-      bufptr += sprintf(bufptr, "l%p", (void *)item_ptr);
+      bufptr += snprintf(bufptr, end-bufptr, "l%p", (void *)item_ptr);
       done = 1;
       break;
     case CLOSXP:
       REPROTECT(item_ptr = item, ix);
-      bufptr += sprintf(bufptr, "c_%p/%p/%p",
+      bufptr += snprintf(bufptr, end-bufptr, "c_%p/%p/%p",
                         (void *) FORMALS(item),
                         (void *) BODY(item),
                         (void *) CLOENV(item));
@@ -164,7 +167,7 @@ SEXP stringify_item(SEXP item, char *bufptr) {
     case NILSXP:
       /* We have an expression-ish, represent its pointer. */
       REPROTECT(item_ptr = item, ix);
-      bufptr += sprintf(bufptr, "e%p", (void *)item_ptr);
+      bufptr += snprintf(bufptr, end-bufptr, "e%p", (void *)item_ptr);
       done = 1;
       break;
     default:
